@@ -13,31 +13,30 @@
 API For Web Map Service version 1.3.0.
 """
 
-from __future__ import (absolute_import, division, print_function)
-
-try:                    # Python 3
-    from urllib.parse import urlencode
-except ImportError:     # Python 2
-    from urllib import urlencode
+from urllib.parse import urlencode
 
 import warnings
-import six
+from math import sqrt
 from owslib.etree import etree
 from owslib.util import (openURL, ServiceException, testXMLValue,
                          extract_xml_list, xmltag_split, OrderedDict, nspath,
-                         bind_url)
-from owslib.util import nspath
+                         nspath_eval, bind_url, Authentication)
 from owslib.fgdc import Metadata
 from owslib.iso import MD_Metadata
 from owslib.crs import Crs
 from owslib.namespaces import Namespaces
-from owslib.map.common import WMSCapabilitiesReader
+from owslib.map.common import WMSCapabilitiesReader, AbstractContentMetadata
 
 from owslib.util import log
 
 n = Namespaces()
 WMS_NAMESPACE = n.get_namespace("wms")
 OGC_NAMESPACE = n.get_namespace('ogc')
+INCH_TO_M = 0.0254  # Standard inch to meter conversion
+OGC_PIXEL_SIZE = 0.00028  # The OGC standard pixel size is 0.28 mm, here it's in meter
+OGC_DPI = INCH_TO_M / OGC_PIXEL_SIZE
+# The sqrt(2) is because the scale is on the diagonal and the resolution on the side
+SCALEDENOM_TO_RESOLUTION = 1 / OGC_DPI * INCH_TO_M * sqrt(2)
 
 
 class WebMapService_1_3_0(object):
@@ -53,20 +52,23 @@ class WebMapService_1_3_0(object):
 
     def __init__(self, url, version='1.3.0', xml=None, username=None,
                  password=None, parse_remote_metadata=False, timeout=30,
-                 headers=None):
+                 headers=None, auth=None):
         """initialize"""
+        if auth:
+            if username:
+                auth.username = username
+            if password:
+                auth.password = password
         self.url = url
-        self.username = username
-        self.password = password
         self.version = version
         self.timeout = timeout
         self.headers = headers
         self._capabilities = None
+        self.auth = auth or Authentication(username, password)
 
         # Authentication handled by Reader
-        reader = WMSCapabilitiesReader(self.version, url=self.url,
-                                       un=self.username, pw=self.password,
-                                       headers=headers)
+        reader = WMSCapabilitiesReader(
+            self.version, url=self.url, headers=headers, auth=self.auth)
         if xml:  # read from stored xml
             self._capabilities = reader.readString(xml)
         else:  # read from server
@@ -113,7 +115,7 @@ class WebMapService_1_3_0(object):
         def gather_layers(parent_elem, parent_metadata):
             layers = []
             for index, elem in enumerate(parent_elem.findall(nspath('Layer', WMS_NAMESPACE))):
-                cm = ContentMetadata(elem, parent=parent_metadata, index=index+1,
+                cm = ContentMetadata(elem, parent=parent_metadata, index=index + 1,
                                      parse_remote_metadata=parse_remote_metadata)
                 if cm.id:
                     if cm.id in self.contents:
@@ -150,9 +152,9 @@ class WebMapService_1_3_0(object):
         raise KeyError("No operation named %s" % name)
 
     def __build_getmap_request(self, layers=None, styles=None, srs=None, bbox=None,
-               format=None, size=None, time=None, dimensions={},
-               elevation=None, transparent=False,
-               bgcolor=None, exceptions=None, **kwargs):
+                               format=None, size=None, time=None, dimensions={},
+                               elevation=None, transparent=False,
+                               bgcolor=None, exceptions=None, **kwargs):
 
         request = {'service': 'WMS', 'version': self.version, 'request': 'GetMap'}
 
@@ -184,9 +186,11 @@ class WebMapService_1_3_0(object):
         request['bbox'] = ','.join([repr(x) for x in bbox])
         request['format'] = str(format)
         request['transparent'] = str(transparent).upper()
-        request['bgcolor'] = '0x' + bgcolor[1:7]
         request['exceptions'] = str(exceptions)
 
+        if bgcolor:
+            request['bgcolor'] = '0x' + bgcolor[1:7]
+
         if time is not None:
             request['time'] = str(time)
 
@@ -194,12 +198,12 @@ class WebMapService_1_3_0(object):
             request['elevation'] = str(elevation)
 
         # any other specified dimension, prefixed with "dim_"
-        for k, v in six.iteritems(dimensions):
+        for k, v in list(dimensions.items()):
             request['dim_' + k] = str(v)
 
         if kwargs:
             for kw in kwargs:
-                request[kw]=kwargs[kw]
+                request[kw] = kwargs[kw]
         return request
 
     def getmap(self, layers=None,
@@ -303,23 +307,18 @@ class WebMapService_1_3_0(object):
 
         self.request = bind_url(base_url) + data
 
-        u = openURL(base_url,
-                    data,
-                    method,
-                    username=self.username,
-                    password=self.password,
-                    timeout=timeout or self.timeout)
+        u = openURL(base_url, data, method, timeout=timeout or self.timeout, auth=self.auth, headers=self.headers)
 
         # need to handle casing in the header keys
         headers = {}
-        for k, v in six.iteritems(u.info()):
+        for k, v in list(u.info().items()):
             headers[k.lower()] = v
 
         # handle the potential charset def
-        if headers['content-type'].split(';')[0] in ['application/vnd.ogc.se_xml', 'text/xml']:
+        if headers.get('content-type', '').split(';')[0] in ['application/vnd.ogc.se_xml', 'text/xml']:
             se_xml = u.read()
             se_tree = etree.fromstring(se_xml)
-            err_message = six.text_type(se_tree.find(nspath('ServiceException', OGC_NAMESPACE)).text).strip()
+            err_message = str(se_tree.find(nspath('ServiceException', OGC_NAMESPACE)).text).strip()
             raise ServiceException(err_message)
         return u
 
@@ -344,7 +343,8 @@ class WebMapService_1_3_0(object):
                        **kwargs
                        ):
         try:
-            base_url = next((m.get('url') for m in self.getOperationByName('GetFeatureInfo').methods if m.get('type').lower() == method.lower()))
+            base_url = next((m.get('url') for m in self.getOperationByName('GetFeatureInfo').methods
+                            if m.get('type').lower() == method.lower()))
         except StopIteration:
             base_url = self.url
 
@@ -362,7 +362,7 @@ class WebMapService_1_3_0(object):
             transparent=transparent,
             bgcolor=bgcolor,
             exceptions=exceptions,
-            kwargs=kwargs)
+            **kwargs)
 
         # extend to GetFeatureInfo-Request
         request['request'] = 'GetFeatureInfo'
@@ -379,19 +379,20 @@ class WebMapService_1_3_0(object):
         request['feature_count'] = str(feature_count)
 
         data = urlencode(request)
- 
+
         self.request = bind_url(base_url) + data
 
-        u = openURL(base_url, data, method, username=self.username, password=self.password, timeout=timeout or self.timeout)
+        u = openURL(base_url, data, method, timeout=timeout or self.timeout, auth=self.auth, headers=self.headers)
 
         # check for service exceptions, and return
         if u.info()['Content-Type'] == 'XML':
             se_xml = u.read()
             se_tree = etree.fromstring(se_xml)
-            err_message = six.text_type(se_tree.find('ServiceException').text).strip()
+            err_message = str(se_tree.find('ServiceException').text).strip()
             raise ServiceException(err_message)
         return u
 
+
 class ServiceIdentification(object):
     def __init__(self, infoset, version):
         self._root = infoset
@@ -412,7 +413,10 @@ class ServiceProvider(object):
             self.name = name.text
         else:
             self.name = None
-        self.url = self._root.find(nspath('OnlineResource', WMS_NAMESPACE)).attrib.get('{http://www.w3.org/1999/xlink}href', '')
+        self.url = None
+        online_resource = self._root.find(nspath('OnlineResource', WMS_NAMESPACE))
+        if online_resource is not None:
+            self.url = online_resource.attrib.get('{http://www.w3.org/1999/xlink}href', '')
         # contact metadata
         contact = self._root.find(nspath('ContactInformation', WMS_NAMESPACE))
         # sometimes there is a contact block that is empty, so make
@@ -423,8 +427,12 @@ class ServiceProvider(object):
             self.contact = None
 
 
-class ContentMetadata(object):
-    def __init__(self, elem, parent=None, children=None, index=0, parse_remote_metadata=False, timeout=30):
+class ContentMetadata(AbstractContentMetadata):
+
+    def __init__(self, elem, parent=None, children=None, index=0, parse_remote_metadata=False,
+                 timeout=30, auth=None):
+        super(ContentMetadata, self).__init__(auth)
+
         if xmltag_split(elem.tag) != 'Layer':
             raise ValueError('%s should be a Layer' % (elem,))
 
@@ -455,8 +463,10 @@ class ContentMetadata(object):
         self.abstract = testXMLValue(elem.find(nspath('Abstract', WMS_NAMESPACE)))
 
         # TODO: what is the preferred response to esri's handling of custom projections
-        #       in the spatial ref definitions? see http://resources.arcgis.com/en/help/main/10.1/index.html#//00sq000000m1000000
-        #       and an example (20150812) http://maps.ngdc.noaa.gov/arcgis/services/firedetects/MapServer/WMSServer?request=GetCapabilities&service=WMS
+        # in the spatial ref definitions? see
+        # http://resources.arcgis.com/en/help/main/10.1/index.html#//00sq000000m1000000
+        # and an example (20150812)
+        # http://maps.ngdc.noaa.gov/arcgis/services/firedetects/MapServer/WMSServer?request=GetCapabilities&service=WMS
 
         # bboxes
         b = elem.find(nspath('EX_GeographicBoundingBox', WMS_NAMESPACE))
@@ -505,11 +515,15 @@ class ContentMetadata(object):
         self.boundingBox = crs_list[0] if crs_list else self.boundingBoxWGS84
 
         # ScaleHint
-        sh = elem.find(nspath('ScaleHint', WMS_NAMESPACE))
         self.scaleHint = None
-        if sh is not None:
-            if 'min' in sh.attrib and 'max' in sh.attrib:
-                self.scaleHint = {'min': sh.attrib['min'], 'max': sh.attrib['max']}
+        self.min_scale_denominator = elem.find(nspath('MinScaleDenominator', WMS_NAMESPACE))
+        min_scale_hint = 0 if self.min_scale_denominator is None else \
+            float(self.min_scale_denominator.text) * SCALEDENOM_TO_RESOLUTION
+        self.max_scale_denominator = elem.find(nspath('MaxScaleDenominator', WMS_NAMESPACE))
+        max_scale_hint = 0 if self.max_scale_denominator is None else \
+            float(self.max_scale_denominator.text) * SCALEDENOM_TO_RESOLUTION
+        if self.min_scale_denominator is not None or self.max_scale_denominator is not None:
+            self.scaleHint = {'min': min_scale_hint, 'max': max_scale_hint}
 
         attribution = elem.find(nspath('Attribution', WMS_NAMESPACE))
         if attribution is not None:
@@ -523,7 +537,8 @@ class ContentMetadata(object):
                 self.attribution['url'] = url.attrib['{http://www.w3.org/1999/xlink}href']
             if logo is not None:
                 self.attribution['logo_size'] = (int(logo.attrib['width']), int(logo.attrib['height']))
-                self.attribution['logo_url'] = logo.find(nspath('OnlineResource', WMS_NAMESPACE)).attrib['{http://www.w3.org/1999/xlink}href']
+                self.attribution['logo_url'] = logo.find(
+                    nspath('OnlineResource', WMS_NAMESPACE)).attrib['{http://www.w3.org/1999/xlink}href']
 
         # TODO: get this from the bbox attributes instead (deal with parents)
         # SRS options
@@ -538,7 +553,7 @@ class ContentMetadata(object):
             # some servers found in the wild use a single SRS
             # tag containing a whitespace separated list of SRIDs
             # instead of several SRS tags. hence the inner loop
-            for srslist in map(lambda x: x.text, elem.findall(nspath('CRS', WMS_NAMESPACE))):
+            for srslist in [x.text for x in elem.findall(nspath('CRS', WMS_NAMESPACE))]:
                 if srslist:
                     for srs in srslist.split():
                         self.crsOptions.append(srs)
@@ -550,7 +565,8 @@ class ContentMetadata(object):
         if len(self.crsOptions) == 0:
             # raise ValueError('%s no SRS available!?' % (elem,))
             # Comment by D Lowe.
-            # Do not raise ValueError as it is possible that a layer is purely a parent layer and does not have SRS specified. Instead set crsOptions to None
+            # Do not raise ValueError as it is possible that a layer is purely a parent layer and
+            # does not have SRS specified. Instead set crsOptions to None
             # Comment by Jachym:
             # Do not set it to None, but to [], which will make the code
             # work further. Fixed by anthonybaxter
@@ -567,9 +583,13 @@ class ContentMetadata(object):
         for s in elem.findall(nspath('Style', WMS_NAMESPACE)):
             name = s.find(nspath('Name', WMS_NAMESPACE))
             title = s.find(nspath('Title', WMS_NAMESPACE))
+            if name is None and title is None:
+                raise ValueError('%s missing name and title' % (s,))
             if name is None or title is None:
-                raise ValueError('%s missing name or title' % (s,))
-            style = {'title': title.text}
+                warnings.warn('%s missing name or title' % (s,))
+            title_ = title.text if title is not None else name.text
+            name_ = name.text if name is not None else title.text
+            style = {'title': title_}
             # legend url
             legend = s.find(nspath('LegendURL/OnlineResource', WMS_NAMESPACE))
             if legend is not None:
@@ -577,15 +597,15 @@ class ContentMetadata(object):
 
             lgd = s.find(nspath('LegendURL', WMS_NAMESPACE))
             if lgd is not None:
-                if 'width' in lgd.attrib.keys():
+                if 'width' in list(lgd.attrib.keys()):
                     style['legend_width'] = lgd.attrib.get('width')
-                if 'height' in lgd.attrib.keys():
+                if 'height' in list(lgd.attrib.keys()):
                     style['legend_height'] = lgd.attrib.get('height')
 
                 lgd_format = lgd.find(nspath('Format', WMS_NAMESPACE))
                 if lgd_format is not None:
                     style['legend_format'] = lgd_format.text.strip()
-            self.styles[name.text] = style
+            self.styles[name_] = style
 
         # keywords
         self.keywords = [f.text for f in elem.findall(nspath('KeywordList/Keyword', WMS_NAMESPACE))]
@@ -622,11 +642,13 @@ class ContentMetadata(object):
         for dim in elem.findall(nspath('Dimension', WMS_NAMESPACE)):
             dim_name = dim.attrib.get('name')
             dim_data = {}
-            for k, v in six.iteritems(dim.attrib):
+            for k, v in list(dim.attrib.items()):
                 if k != 'name':
                     dim_data[k] = v
             # single values and ranges are not differentiated here
-            dim_data['values'] = dim.text.strip().split(',') if dim.text.strip() else None
+            dim_data['values'] = None
+            if dim.text:
+                dim_data['values'] = dim.text.strip().split(',') if dim.text.strip() else None
             self.dimensions[dim_name] = dim_data
 
         # MetadataURLs
@@ -635,30 +657,23 @@ class ContentMetadata(object):
             metadataUrl = {
                 'type': testXMLValue(m.attrib['type'], attrib=True),
                 'format': testXMLValue(m.find(nspath('Format', WMS_NAMESPACE))),
-                'url': testXMLValue(m.find(nspath('OnlineResource', WMS_NAMESPACE)).attrib['{http://www.w3.org/1999/xlink}href'], attrib=True)
+                'url': testXMLValue(m.find(
+                    nspath('OnlineResource', WMS_NAMESPACE)).attrib['{http://www.w3.org/1999/xlink}href'], attrib=True)
             }
-
-            if metadataUrl['url'] is not None and parse_remote_metadata:  # download URL
-                try:
-                    content = openURL(metadataUrl['url'], timeout=timeout)
-                    doc = etree.parse(content)
-                    if metadataUrl['type'] is not None:
-                        if metadataUrl['type'] == 'FGDC':
-                            metadataUrl['metadata'] = Metadata(doc)
-                        if metadataUrl['type'] == 'TC211':
-                            metadataUrl['metadata'] = MD_Metadata(doc)
-                except Exception:
-                    metadataUrl['metadata'] = None
-
             self.metadataUrls.append(metadataUrl)
 
+        if parse_remote_metadata:
+            self.parse_remote_metadata(timeout)
+
         # DataURLs
         self.dataUrls = []
         for m in elem.findall(nspath('DataURL', WMS_NAMESPACE)):
             dataUrl = {
-                'format': m.find(nspath('Format', WMS_NAMESPACE)).text.strip(),
+                'format': m.find(nspath('Format', WMS_NAMESPACE)).text,
                 'url': m.find(nspath('OnlineResource', WMS_NAMESPACE)).attrib['{http://www.w3.org/1999/xlink}href']
             }
+            if dataUrl['format'] is not None:
+                dataUrl['format'] = dataUrl['format'].strip()
             self.dataUrls.append(dataUrl)
 
         # FeatureListURLs
@@ -674,6 +689,29 @@ class ContentMetadata(object):
         for child in elem.findall(nspath('Layer', WMS_NAMESPACE)):
             self.layers.append(ContentMetadata(child, self))
 
+    def parse_remote_metadata(self, timeout=30):
+        """Parse remote metadata for MetadataURL and add it as metadataUrl['metadata']"""
+        for metadataUrl in self.metadataUrls:
+            if metadataUrl['url'] is not None \
+                    and metadataUrl['format'].lower() in ['application/xml', 'text/xml']:  # download URL
+                try:
+                    content = openURL(
+                        metadataUrl['url'], timeout=timeout, auth=self.auth)
+                    doc = etree.fromstring(content.read())
+
+                    mdelem = doc.find('.//metadata')
+                    if mdelem is not None:
+                        metadataUrl['metadata'] = Metadata(mdelem)
+                        continue
+
+                    mdelem = doc.find('.//' + nspath_eval('gmd:MD_Metadata', n.get_namespaces(['gmd']))) \
+                        or doc.find('.//' + nspath_eval('gmi:MI_Metadata', n.get_namespaces(['gmi'])))
+                    if mdelem is not None:
+                        metadataUrl['metadata'] = MD_Metadata(mdelem)
+                        continue
+                except Exception:
+                    metadataUrl['metadata'] = None
+
     @property
     def children(self):
         return self._children
@@ -688,6 +726,7 @@ class ContentMetadata(object):
     def __str__(self):
         return 'Layer Name: %s Title: %s' % (self.name, self.title)
 
+
 class OperationMetadata(object):
     def __init__(self, elem):
         """."""
